/**
 * $Id: ClasspathMonitor.java,v 1.14 2009/03/09 21:18:54 oboehm Exp $
 *
 * Copyright (c) 2008 by Oliver Boehm
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express orimplied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 * (c)reated 10.02.2009 by oliver (ob@oasd.de)
 */

package patterntesting.runtime.monitor;

import java.io.*;
import java.lang.reflect.*;
import java.net.*;
import java.util.*;
import java.util.concurrent.TimeUnit;
import java.util.jar.JarFile;

import javax.management.JMException;

import org.apache.commons.lang.StringUtils;
import org.apache.log4j.Logger;

import patterntesting.runtime.annotation.ProfileMe;
import patterntesting.runtime.jmx.MBeanRegistry;
import patterntesting.runtime.util.*;

/**
 * To avoid classpath problems like double entries of the same class or resource
 * in the classpath there are several methods available.
 * 
 * @author <a href="boehm@javatux.de">oliver</a>
 * @since 10.02.2009
 * @version $Revision: 1.14 $
 * @see http://www.javaworld.com/javaworld/javatips/jw-javatip105.html
 */
public class ClasspathMonitor implements ClasspathMonitorMBean, MBeanRegistry {

    private static final long serialVersionUID = 20090212L;
    private static final Logger log = Logger.getLogger(ClasspathMonitor.class);
    private final ClassLoader cloader;
    private int nLoadedClasses = -1;
    private List<Class<?>> doubletList = new ArrayList<Class<?>>();
    private SortedSet<URI> usedClasspathSet = new TreeSet<URI>();

    public ClasspathMonitor() {
        this.cloader = checkClassloader(this.getClass().getClassLoader());
    }

    public ClasspathMonitor(ClassLoader loader) {
        this.cloader = checkClassloader(loader);
    }
    
    private void resetCache() {
        doubletList = new ArrayList<Class<?>>();
        usedClasspathSet = new TreeSet<URI>();
    }

    private ClassLoader checkClassloader(ClassLoader classloader) {
        if (classloader == null) {
            classloader = Thread.currentThread().getContextClassLoader();
            log.warn("null ClassLoader given - replaced by " + classloader);
        }
        return classloader;
    }

    /**
     * @see ClasspathMonitorMBean#whichResource(String)
     */
    @ProfileMe
    public URI whichResource(String name) {
        URL classUrl = ClasspathMonitor.class.getResource(name);
        if (classUrl == null) {
            if (!name.startsWith("/")) {
                return whichResource("/" + name);
            }
            if (log.isTraceEnabled()) {
            	log.trace(name + " not found in classpath");
            }
            return null;
        } else {
            if (log.isTraceEnabled()) {
                log.trace("using " + classUrl + " for " + name);
            }
        }
        try {
        	return classUrl.toURI();
        } catch (URISyntaxException canthappen) {
            String msg = "can't convert " + classUrl + " to URI";
            log.error(msg, canthappen);
            throw new RuntimeException(canthappen);
        }
    }

    /**
     * Returns the URI of the given resource and the given classloader.
     * 
     * @param name
     *            resource name (e.g. "log4j.properties")
     * @param classloader
     * @return URI of the given resource (or null if resource was not found)
     */
    @ProfileMe
    public URI whichResource(String name, ClassLoader classloader) {
        assert classloader != null : "no classloader given";
        URL url = classloader.getResource(name);
        if (url == null) {
            if (name.startsWith("/")) {
                return whichResource(name.substring(1));
            }
            return null;
        }
        try {
            return new URI(url.toString());
        } catch (URISyntaxException canthappen) {
            String msg = "can't convert " + url + " to URI";
            log.error(msg, canthappen);
            throw new RuntimeException(canthappen);
        }
    }

    /**
     * @see patterntesting.runtime.monitor.ClasspathMonitorMBean#whichClass(java.lang.String)
     */
    @ProfileMe
    public URI whichClass(String name) {
        String resource = Converter.classToResource(name);
        return whichResource(resource);
    }
    
    /**
     * Where is the given class found in the classpath? Which class was loaded
     * from which URI?
     * 
     * @param clazz
     *            the class
     * @return file or jar path
     */
    @ProfileMe
    public URI whichClass(Class<?> clazz) {
        return whichClass(clazz.getName());
    }
    
    /**
     * Returns the jar file or path where the given classname was found
     * 
     * @param classname e.g. java.lang.String
     * @return jar or path as URI
     */
    public URI whichClassPath(String classname) {
        String resource = Converter.classToResource(classname);
        return whichResourcePath(resource);
    }
    
    /**
     * Returns the jar file or path where the given resource was found
     * 
     * @param resource e.g. log4j.properties
     * @return jar or path as URI
     */
    public URI whichResourcePath(String resource) {
        URI uri = this.whichResource(resource);
        if (uri == null) {
            return null;
        }
        return getParent(uri, resource);
    }
    
    public JarFile whichClassJar(Class<?> clazz) {
        return whichClassJar(clazz.getName());
    }
    
    public JarFile whichClassJar(String classname) {
        String resource = Converter.classToResource(classname);
        return whichResourceJar(resource);
    }
    
    public JarFile whichResourceJar(String resource) {
        return whichResourceJar(this.whichResourcePath(resource));
    }
    
    public static JarFile whichResourceJar(URI resource) {
        if (resource == null) {
            return null;
        }
        File file = Converter.toFile(resource);
        try {
            return new JarFile(file);
        } catch (IOException ioe) {
            log.debug("can't read " + file, ioe);
            return null;
        }        
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * patterntesting.runtime.ClasspathCheckerMBean#getResources(java.lang.String
     * )
     */
    @ProfileMe
    public Enumeration<URL> getResources(String name) {
        try {
            Enumeration<URL> resources = cloader.getResources(name);
            if (!resources.hasMoreElements()) {
                if (name.startsWith("/")) {
                    return getResources(name.substring(1));
                }
                if (log.isDebugEnabled()) {
                    log.debug(name + " not found in classpath");
                }
            }
            return resources;
        } catch (IOException ioe) {
            log.info(name + " not found in classpath", ioe);
            return null;
        }
    }

    public static URL getResource(String name) {
        return ClasspathMonitor.class.getResource(name);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * patterntesting.runtime.ClasspathCheckerMBean#getNoResources(java.lang
     * .String)
     */
    @ProfileMe
    public int getNoResources(String name) {
        Enumeration<URL> resources = getResources(name);
        if (resources == null) {
            return 0;
        }
        int n = 0;
        while (resources.hasMoreElements()) {
            n++;
            resources.nextElement();
        }
        if (log.isTraceEnabled()) {
        	log.trace(n + " element(s) of " + name + " found in classpath");
        }
        return n;
    }

    public int getNoClasses(Class<?> cl) {
        return getNoClasses(cl.getName());
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * patterntesting.runtime.ClasspathCheckerMBean#getNoClasses(java.lang.String
     * )
     */
    public int getNoClasses(String classname) {
        return getNoResources(Converter.classToResource(classname));
    }

    /**
     * @see patterntesting.runtime.monitor.ClasspathMonitorMBean#isDoublet(java.lang.String)
     */
    @ProfileMe
    public boolean isDoublet(String name) {
        Enumeration<URL> resources = getResources(name);
        if ((resources == null) || !resources.hasMoreElements()) {
            throw new NoSuchElementException(name);
        }
        resources.nextElement();
        if (resources.hasMoreElements()) {
            logDoublets(name);
            return true;
        } else {
            return false;
        }
    }

    /**
     * Is the given class a doublet, i.e. can it be found more than once in the
     * classpath?
     * 
     * @param clazz
     *            the class to be checked
     * @return true if class is found more than once in the classpath
     * @throws NoSuchElementException
     *             if class is not found in the classpath
     */
    @ProfileMe
    public boolean isDoublet(Class<?> clazz) {
        String classname = clazz.getName();
        String resource = Converter.classToResource(classname);
        return isDoublet(resource);
    }

    /**
     * @see patterntesting.runtime.monitor.ClasspathMonitorMBean#getFirstDoublet(java.lang
     *      .String)
     */
    public URL getFirstDoublet(String name) {
        return getDoublet(name, 1);
    }

    /**
     * @see patterntesting.runtime.monitor.ClasspathMonitorMBean#getDoublet(java.lang.String,
     *      int)
     */
    public URL getDoublet(String name, int nr) {
        Enumeration<URL> resources = getResources(name);
        for (int i = 0; resources.hasMoreElements(); i++) {
            URL url = (URL) resources.nextElement();
            if (i == nr) {
                return url;
            }
        }
        return null;
    }

    public URL getDoublet(Class<?> clazz, int nr) {
        String resource = Converter.classToResource(clazz.getName());
        return getDoublet(resource, nr);
    }

    private void logDoublets(String name) {
        if (log.isTraceEnabled()) {
            List<URL> doublets = new Vector<URL>();
            Enumeration<URL> resources = getResources(name);
            while (resources.hasMoreElements()) {
                URL url = resources.nextElement();
                doublets.add(url);
            }
            log.trace(name + " doublets found: " + doublets);
        }
    }

    /**
     * Looks for each loaded class if it is a doublet or not.
     * For the performance reason this is only done if the number of loaded
     * classes has changed since the last call.
     * 
     * @return a sorted list of found doublets
     */
    @ProfileMe
    public synchronized List<Class<?>> getDoubletList() {
        List<Class<?>> loadedClassList = this.getLoadedClassList();
        if (!doubletList.isEmpty()) {
            return doubletList;
        }
        for (Class<?> clazz : loadedClassList) {
            try {
                if (this.isDoublet(clazz)) {
                    doubletList.add(clazz);
                }
            } catch (NoSuchElementException nsee) {
                if (log.isTraceEnabled()) {
                    log.trace(clazz + " not found -> ignored");
                }
            }
        }
        Collections.sort(doubletList, new ObjectComparator());
        return doubletList;
    }
    
    /**
     * @see ClasspathMonitorMBean#getDoublets()
     */
    @ProfileMe
    public String[] getDoublets() {
        List<Class<?>> doubletList = this.getDoubletList();
        String[] doublets = new String[doubletList.size()];
        for (int i = 0; i < doublets.length; i++) {
            doublets[i] = doubletList.get(i).toString();
        }
        return doublets;
    }
    
    /**
     * @see ClasspathMonitorMBean#getDoubletClasspath()
     */
    @ProfileMe
    public String[] getDoubletClasspath() {
    	Set<URI> classpathSet = getDoubletClasspathSet();
        return getAsArray(classpathSet);
    }

    @ProfileMe
    public SortedSet<URI> getDoubletClasspathSet() {
    	List<Class<?>> doubletList = this.getDoubletList();
    	SortedSet<URI> classpath = new TreeSet<URI>();
    	for (Class<?> cl : doubletList) {
			String resource = Converter.classToResource(cl);
			Enumeration<URL> resources = this.getResources(resource);
			while(resources.hasMoreElements()) {
				URL url = resources.nextElement();
				try {
					URI path = url.toURI();
					classpath.add(getParent(path, resource));
				} catch (URISyntaxException canthappen) {
					log.debug("can't convert " + url + " to URI", canthappen);
				}
			}
		}
    	return classpath;
    }

    /**
     * Dumps the internal fields of the classloader (after an idea from
     * "Java ist auch eine Insel").
     * 
     * @see http 
     *      ://www.galileocomputing.de/openbook/javainsel6/javainsel_21_003.htm
     *      #mj7e2d89f4b96fde1900bc09fd1db83fb1
     */
    @ProfileMe
    public String getClassLoaderDetails() {
        StringBuffer sbuf = new StringBuffer("dump of " + cloader + ":\n");
        for (Class<?> cl = this.cloader.getClass(); cl != null; cl = cl
                .getSuperclass()) {
            sbuf.append('\t');
            dumpFields(sbuf, cl, this.cloader);
        }
        return sbuf.toString().trim();
    }

    @ProfileMe
    private static final void dumpFields(StringBuffer sbuf, Class<?> cl,
            Object obj) {
        Field[] fields = cl.getDeclaredFields();
        AccessibleObject.setAccessible(fields, true);
        for (int i = 0; i < fields.length; i++) {
            sbuf.append(fields[i]);
            sbuf.append(" = ");
            try {
                sbuf.append(fields[i].get(obj));
            } catch (Exception e) {
                sbuf.append("<" + e + ">");
            }
            sbuf.append('\n');
        }
    }

    /**
     * Returns the packages which were loaded by the classloader. <br/>
     * Unfortunately ClassLoader.getPackages() is protected - ok, let's do the
     * hard way using reflexion.
     * 
     * @return array with the loaded packages
     */
    @ProfileMe
    public Package[] getLoadedPackageArray() {
        Method method = getMethod(cloader.getClass(), "getPackages");
        if (method == null) {
            log.warn(cloader.getClass() + "#getPackages() not found");
            return null;
        }
        try {
            method.setAccessible(true);
            return (Package[]) method.invoke(cloader);
        } catch (Exception e) {
            log.warn("can't access " + method, e);
            return null;
        }
    }

    /**
     * @see ClasspathMonitorMBean#getLoadedPackages()
     */
    public String[] getLoadedPackages() {
        Package[] packages = getLoadedPackageArray();
        String[] strings = new String[packages.length];
        for (int i = 0; i < packages.length; i++) {
            strings[i] = packages[i].toString();
        }
        Arrays.sort(strings);
        return strings;
    }

    /**
     * If you want to dump all packages you can use this method. The output will
     * be sorted.
     * 
     * @return each package in a single line
     */
    public String getLoadedPackagesAsString() {
        String[] packages = getLoadedPackages();
        StringBuffer sbuf = new StringBuffer();
        for (int i = 0; i < packages.length; i++) {
            sbuf.append(packages[i]);
            sbuf.append('\n');
        }
        return sbuf.toString().trim();
    }

    private static Method getMethod(Class<?> cl, String name) {
        try {
            return cl.getDeclaredMethod(name);
        } catch (SecurityException e) {
            log.warn("can't get method " + cl + "#" + name, e);
            return null;
        } catch (NoSuchMethodException e) {
            Class<?> superclass = cl.getSuperclass();
            if (superclass == null) {
                return null;
            } else {
                return getMethod(superclass, name);
            }
        }
    }

    /**
     * Returns a list of classes which were loaded by the given classloader. <br/>
     * Ok, we must do some hacks here: there is an undocumented attribute
     * "classes" which contains the loaded classes. <br/>
     * HANDLE WITH CARE (it's a hack and it depends on the used classloader)
     * 
     * @param cloader
     *            classloader
     * @return list of classes
     */
    @ProfileMe
    @SuppressWarnings("unchecked")
    private static List<Class<?>> getLoadedClassList(ClassLoader cloader) {
        Field field = ReflectionHelper.getField(cloader.getClass(), "classes");
        field.setAccessible(true);
        try {
            List<Class<?>> classList = (List<Class<?>>) field.get(cloader);
            return new ArrayList<Class<?>>(classList);
        } catch (Exception e) {
            log.warn("can't access " + field, e);
            return Collections.EMPTY_LIST;
        }
    }

    /**
     * Returns a list of classes which were loaded by the classloader.
     * If the number of classes has changed the internal cache for
     * getDoubletList() and getUsedClasspath() will be deleted.
     * 
     * @return list of classes
     */
    @ProfileMe
    public List<Class<?>> getLoadedClassList() {
        List<Class<?>> classlist = getLoadedClassList(cloader);
        int n = classlist.size();
        if (n != this.nLoadedClasses) {
            nLoadedClasses = n;
            resetCache();
        }
        return classlist;
    }

    /**
     * As MBean a string array could be displayed by the 'jconsole'. A class
     * array not.
     * 
     * @return the classes as sorted string array
     */
    public String[] getLoadedClasses() {
        List<Class<?>> classes = getLoadedClassList();
        String[] strings = new String[classes.size()];
        for (int i = 0; i < strings.length; i++) {
            strings[i] = classes.get(i).toString();
        }
        Arrays.sort(strings);
        return strings;
    }

    @ProfileMe
    public String getLoadedClassesAsString() {
        List<Class<?>> classes = getLoadedClassList();
        Collections.sort(classes, new ObjectComparator());
        StringBuffer sbuf = new StringBuffer();
        for (Iterator<Class<?>> i = classes.iterator(); i.hasNext();) {
            sbuf.append(i.next().toString().trim());
            sbuf.append('\n');
        }
        return sbuf.toString().trim();
    }

    /**
     * Gets the loaded classpath (without the bootclasspath) as sorted set.
     * 
     * @return the loaded classpath (excluding the bootclasspath)
     */
    @ProfileMe
    public SortedSet<URI> getUsedClasspathSet() {
        List<Class<?>> loadedClassList = this.getLoadedClassList();
        if (!usedClasspathSet.isEmpty()) {
            return this.usedClasspathSet;
        }
        for (Class<?> clazz : loadedClassList) {
            URI classUri = this.whichClass(clazz);
            if (classUri != null) {
                URI classpath = getParent(classUri, clazz);
                usedClasspathSet.add(classpath);
            }
        }
        return usedClasspathSet;
    }

    /**
     * It might be that the used classpath changes during the calculation of it.
     * So it is only a good approximation how it looks at the moment of the
     * call.
     * 
     * @see ClasspathMonitor#getUsedClasspath()
     */
    @ProfileMe
    public String[] getUsedClasspath() {
        SortedSet<URI> classpathSet = this.getUsedClasspathSet();
        return getAsArray(classpathSet);
    }
    
    private static String[] getAsArray(Set<URI> classpathSet) {
        String[] classpath = new String[classpathSet.size()];
        Object[] a = classpathSet.toArray();
        for (int i = 0; i < classpath.length; i++) {
            URI uri = (URI) a[i];
            classpath[i] = uri.getPath();
            if (StringUtils.isEmpty(classpath[i])) {
                classpath[i] = StringUtils.substringAfterLast(uri.toString(),
                        ":");
            }
            if (classpath[i].endsWith(File.separator)) {
                classpath[i] = classpath[i].substring(0,
                        (classpath[i].length() - 1));
            }
        }
        Arrays.sort(classpath);
        return classpath;
    }
    
    /**
     * The unused classpath is this path which are not used in past.
     * <br/>
     * TODO: the first three calls can be parallized (for better performance)
     * 
     * @return the unused classpath (sorted)
     */
    @ProfileMe
    public String[] getUnusedClasspath() {
        SortedSet<String> unused = new TreeSet<String>();
        String[] classpath = this.getClasspath();
        log.trace(Arrays.class + " loaded (to get corrected used classpath");
        String[] used = this.getUsedClasspath();
        for (int i = 0; i < classpath.length; i++) {
            String path = classpath[i];
            if (Arrays.binarySearch(used, path) < 0) {
                unused.add(path);
            }
        }
        String[] a = new String[unused.size()];
        return unused.toArray(a);
    }

    
    /**
     * To get the boot classpath the sytem property "sun.boot.class.path" is
     * used to get them. This will work of course only for the SunVM.
     * 
     * @return the boot classpath as String array
     */
    public String[] getBootClasspath() {
        String classpath = System.getProperty("sun.boot.class.path");
        if (classpath == null) {
            log.info("sun.boot.class.path is not set (not a SunVM?)");
            return new String[0];
        }
        return splitClasspath(classpath);
    }
    
    /**
     * The classpath is stored in the system property "java.class.path".
     * And this is the classpath which will be returned.
     * 
     * @return the classpath as String array
     */
    public String[] getClasspath() {
        String classpath = System.getProperty("java.class.path");
        return splitClasspath(classpath);
    }
    
    private static String[] splitClasspath(String classpath) {
        String separator = System.getProperty("path.separator", ":");
        String[] cp = StringUtils.split(classpath, separator);
        for (int i = 0; i < cp.length; i++) {
            if (cp[i].endsWith(File.separator)) {
                cp[i] = cp[i].substring(0, (cp[i].length() - 1));
            }
        }
        return cp;
    }

    @ProfileMe
    private static URI getParent(URI path, Class<?> clazz) {
        String resource = Converter.classToResource(clazz);
        return getParent(path, resource);
    }

    /**
     * Don't use @ProfileMe here!
     * This method is use also when ProfileStatistic is set up.
     *
     * @param path e.g. "jar:file:/a/b/c.jar!/d/e.class
     * @param resource e.g. "/d/e.class"
     * @return e.g. "jar:file:/a/b/c.jar"
     */
    public static URI getParent(URI path, String resource) {
        String parent = StringUtils.removeEnd(path.toString(), resource);
        if (parent.endsWith(File.separator)) {
            parent = parent.substring(0, (parent.length() - 1));
        }
        if (parent.endsWith("!")) {
            parent = parent.substring(0, (parent.length() - 1));
        }
        try {
            return new URI(parent);
        } catch (URISyntaxException canthappen) {
            log.error("can't extract " + resource + " from " + path,
                            canthappen);
            return path;
        }
    }

    /**
     * This main method is only for testing the ClasspathMonitor together with
     * the 'jconsole'. On a Java5 VM start it with the following options:
     * -Dcom.sun.management.jmxremote.local.only=false
     * -Dcom.sun.management.jmxremote
     * 
     * @throws JMException
     */
    public static void main(String[] args) throws JMException {
        ClasspathMonitor cpChecker = new ClasspathMonitor();
        cpChecker.registerAsMBean();
        ThreadUtil.sleep(300, TimeUnit.SECONDS);
    }

}

/*
 * $Log: ClasspathMonitor.java,v $
 * Revision 1.14  2009/03/09 21:18:54  oboehm
 * refactored out some parts to ReflectionHelper
 *
 * Revision 1.13  2009/02/27 19:51:24  oboehm
 * some minor optimization in whichResource(String)
 *
 * Revision 1.12  2009/02/26 22:24:19  oboehm
 * cache for getUsedClasspath() added
 *
 * Revision 1.11  2009/02/26 21:48:27  oboehm
 * getDoubletList() tuned
 *
 * Revision 1.10  2009/02/26 10:02:19  oboehm
 * improved JAMon detection (only JAMon 2.4 will be used if found in classpath)
 * output of ProfileStatistic improved (Label contains no longer return value;
 * this is better for sorting in Excel or OpenOffice)
 *
 * Revision 1.9  2009/02/25 21:22:16  oboehm
 * because of problems with JAMon 2.0 only JAMon 2.4 will be supported
 * (if found in the classpath)
 *
 * Revision 1.8  2009/02/21 21:07:45  oboehm
 * getDoubletClasspath() added
 *
 * Revision 1.7  2009/02/20 22:08:34  oboehm
 * some log levels reduced
 *
 * Revision 1.6  2009/02/20 21:45:40  oboehm
 * using MBeanRegistry for registration of ClasspathMonitorMBean
 *
 * Revision 1.5  2009/02/19 12:33:09  oboehm
 * OnlyForTesting support added
 *
 * Revision 1.4  2009/02/18 20:57:30  oboehm
 * getUnusedClasspath() added
 *
 * Revision 1.3  2009/02/18 19:25:40  oboehm
 * getLoadedClasspath() returns now a String array
 *
 * Revision 1.2  2009/02/16 22:14:06  oboehm
 * getBootClasspath() added
 *
 * Revision 1.1  2009/02/15 21:49:46  oboehm
 * ClasspathChecker rename to ClasspathMonitor
 * getLoadedClasspath() added
 * $Source: /cvsroot/patterntesting/PatternTesting08/patterntesting-rt/src/main/java/patterntesting/runtime/monitor/ClasspathMonitor.java,v $
 */
