package patterntesting.runtime.monitor;

import java.io.Serializable;
import java.net.*;

public interface ClasspathMonitorMBean extends Serializable {

    /**
     * Looks if the given resource can be found in the classpath.
     * <br/>
     * To avoid problems like "java.rmi.UnmarshalException: failed to unmarshal
     * class java.lang.Object; nested exception is: java.io.IOException: unknown
     * protocol: zip" no longer a URL but URI is now returned
     * 
     * @param name of a resource e.g. "log4j.properties"
     * @return URI of the given resource (or null if resource was not found)
     */
    public URI whichResource(String name);

    /**
     * Looks if the given classname can be found in the classpath.
     * <br/>
     * To avoid problems like "java.rmi.UnmarshalException: failed to unmarshal
     * class java.lang.Object; nested exception is: java.io.IOException: unknown
     * protocol: zip" no longer a URL but URI is now returned
     * 
     * @param name of a class e.g. "java.lang.String"
     * @return URI of the given resource (or null if resource was not found)
     */
    public URI whichClass(String name);

    /**
     * Get number of resources.
     * 
     * @param name of the resource
     * @return how often the resource was found
     */
    public int getNoResources(String name);

    /**
     * Get the number of different versions of the given class.
     * 
     * @param classname
     * @return how often the classname was found in the classpath
     */
    public int getNoClasses(String classname);

    /**
     * Is the given classname or resource a doublet, e.g. can it be found
     * several times in the classpath?
     * If the classname or resource is not in the classpath a
     * java.util.NoSuchElementException will be thrown.
     * 
     * @param name a classname or resource
     * @return true if more than one classname or resource was found in the
     *         classpath
     * @throws NoSuchElementException if no classname or resource was found
     */
    public boolean isDoublet(String name);

    /**
     * Returns the first doublet of the given classname or resource.
     * 
     * @param name a classname or resource
     * @return
     */
    public URL getFirstDoublet(String name);
    
    /**
     * Looks for each loaded class if it is a doublet or not.
     * 
     * @return a sorted array with the found doublets
     */
    public String[] getDoublets();
    
    /**
     * Looks for each found doublet in which classpath it was found.
     * 
     * @return the classpath where doublets were found
     */
    public String[] getDoubletClasspath();

    /**
     * Returns the n'th doublet of the given classname or resource
     * 
     * @param name a classname or resource
     * @param n
     * @return
     */
    public URL getDoublet(String name, int n);

    /**
     * Returns the packages which were loaded by the classloader.
     * The loaded packages are returned as string array so that it can
     * be displayed by the 'jconsole'.
     * 
     * @return the packages as string array
     */
    public String[] getLoadedPackages();
    
    /**
     * Returns the classes which were loaded by the classloader.
     * The loaded packages are returned as string array so that it can
     * be displayed the 'jconsole'.
     * 
     * @return the classes as string array
     */
    public String[] getLoadedClasses();
    
    /**
     * Gets the loaded classpath (without the bootclasspath) as URI array.
     * 
     * @return the loaded classpath (sorted)
     */
    public String[] getUsedClasspath();
    
    /**
     * The unused classpath is this path which are not used in past.
     * 
     * @return the unused classpath (sorted)
     */
    public String[] getUnusedClasspath();
    
    /**
     * Gets the boot classpath.
     * 
     * @return boot classpath
     */
    public String[] getBootClasspath();
    
    /**
     * Gets the normal classpath which is also available by the
     * system property "java.class.path".
     * 
     * @return the classpath as String array
     */
    public String[] getClasspath();

}

/*
 * $Log: ClasspathMonitorMBean.java,v $
 * Revision 1.5  2009/02/21 21:07:45  oboehm
 * getDoubletClasspath() added
 *
 * Revision 1.4  2009/02/18 20:57:32  oboehm
 * getUnusedClasspath() added
 *
 * Revision 1.3  2009/02/18 19:25:42  oboehm
 * getLoadedClasspath() returns now a String array
 *
 * Revision 1.2  2009/02/16 22:14:06  oboehm
 * getBootClasspath() added
 *
 * Revision 1.1  2009/02/15 21:49:46  oboehm
 * ClasspathChecker rename to ClasspathMonitor
 * getLoadedClasspath() added
 *
 * $Source: /cvsroot/patterntesting/PatternTesting08/patterntesting-rt/src/main/java/patterntesting/runtime/monitor/ClasspathMonitorMBean.java,v $
 */
